from math import ceil
from typing_extensions import List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np


def transform_and_plot(
    c: np.ndarray,
    X: np.ndarray,
    transformers: List,
    title: Optional[str] = None,
    fname: Optional[str] = None,
    xlog: bool = False,
    ylog: bool = False,
) -> Tuple[plt.Figure, np.ndarray]:
    """Creates figure with subplots of transformed data. The first subplot is the original data.
    The rest of the subplots are the transformed data.

    Args:
      -  c (np.ndarray): The x-axis values.
      -  X (np.ndarray): The y-axis values.
      -  transformers (List): List of transformers.
      -  title (Optional[str], optional): Title of the figure. Defaults to None.
      -  fname (Optional[str], optional): File name to save the figure. Defaults to None.
      -  xlog (bool, optional): If True, the x-axis is logarithmic. Defaults to False.
      -  ylog (bool, optional): If True, the y-axis is logarithmic. Defaults to False.

    Returns:
       - Tuple[plt.Figure, np.ndarray]: The figure and the axes.
    """
    if not isinstance(transformers, (list, tuple)):
        transformers = [transformers]
    trlen = len(transformers)
    if not isinstance(xlog, (list, tuple)):
        xlog = [xlog] * (trlen + 1)
    if not isinstance(ylog, (list, tuple)):
        ylog = [ylog] * (trlen + 1)
    if len(xlog) != trlen + 1:
        raise ValueError("xlog and transformers length mismatch")
    if len(ylog) != trlen + 1:
        raise ValueError("ylog and transformers length mismatch")
    n = len(transformers)
    # calculate number of rows if maximum number of columns is 3
    nrows = 1 if n < 3 else ceil((n + 1) / 3)
    ncols = 3 if n >= 3 else n + 1
    fig, axes = plt.subplots(
        nrows, ncols, figsize=(4 * ncols, 4 * nrows), squeeze=False
    )
    axes = axes.flatten()
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    for ax in axes[1:]:
        ax.set_visible(False)
    if xlog[0]:
        axes[0].set_xscale("log")
    if ylog[0]:
        axes[0].set_yscale("log")

    Xt = X
    ct = c
    Xt_ = Xt.T if X.ndim == 2 else Xt
    ct_ = ct.T if c.ndim == 2 else ct

    axes[0].plot(ct_, Xt_, "-")
    axes[0].set_title("Original data")

    for i, transformer in enumerate(transformers):
        axes[i + 1].set_visible(True)
        Xt = transformer(Xt, c)
        if xlog[i + 1]:
            axes[i + 1].set_xscale("log")
        if ylog[i + 1]:
            axes[i + 1].set_yscale("log")
        Xt_ = Xt.T if X.ndim == 2 else Xt
        ct_ = ct.T if c.ndim == 2 else ct
        axes[i + 1].plot(ct_, Xt_, "-")
        axes[i + 1].set_title(transformer.__class__.__name__)
    if title is not None:
        fig.suptitle(title)
    if fname is not None:
        fig.savefig(fname)
    return fig, axes
