from typing_extensions import Type, Union

from escape.core.setting import int_setting_obj
from typeguard import typechecked

from escape.core.model import model_obj, modelstack, modelstack_obj

from .mlp import MLP
from .module import Module
from .regressor import Regressor, loads_torch_regressor


@typechecked
def torch_regressor(
    name,
    mstack: Union[modelstack_obj, list[model_obj], model_obj],
    *,
    module: Union[Type[Module], Module] = MLP,
    nsamples: int = 1000,
    ntests: int = 100,
    data_noisifiers: Union[callable, list[callable]] = None,
    coordinates_noisifiers: Union[callable, list[callable]] = None,
    data_normalizers: Union[callable, list[callable]] = None,
    **kwargs,
):
    """
    Creates torch regressor object.

    :Parameters:
        `name`: string
            Name of regressor object.
        `mstack`: modelstack_obj, list of models or model_obj
            Models to be optimized.
        `module`: Python class or instance of `Module`
            Name of module class or instance of `Module` to be used.
        `nsamples`: positive integer
            Number of training datasets to be generated.
        `ntests`: positive integer
            Number of test datasets to be generated.
        `data_noisifiers`: callable or list of callables
            Noisifier class instance to be used to noisify generated training data.
        `coordinates_noisifiers`: callable or list of callables
            Instance to be used to noisify generated training data coordinates.
        `data_normalizers`: callable or list of callables
            Method or instance to be used to normalize generated training data.
        `kwargs`: dict
            Keyword arguments for regressor.

    :return:
        instance of regressor.torch.Regressor

    """
    nsamples_: int_setting_obj = int_setting_obj.convert(nsamples)
    ntests_: int_setting_obj = int_setting_obj.convert(ntests)
    stack: modelstack_obj = modelstack(f"Modelstack: {name}", mstack)
    resobj: Regressor = Regressor(
        name,
        stack,
        module=module,
        nsamples=nsamples_,
        ntests=ntests_,
        data_noisifiers=data_noisifiers,
        coordinates_noisifiers=coordinates_noisifiers,
        data_normalizers=data_normalizers,
        **kwargs,
    )
    return resobj
