from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

import numpy as np
from escape.utils.widgets.base import BaseLayoutWithPlot, LayoutConfig, unique
from escape.utils.widgets.plot import (
    MapPlotConfig,
    MapWidget,
    Plot2dConfig,
    Plot2dWidget,
)
from numpy.typing import ArrayLike
from typeguard import typechecked

if TYPE_CHECKING:
    from escape.core.kernel import kernel_obj


@dataclass(frozen=True)
class KernelConfig(LayoutConfig, Plot2dConfig, MapPlotConfig):
    """Configuration for kernel object layout."""


class KernelObjectLayout(BaseLayoutWithPlot):
    """Layout for displaying kernel objects with configurable coordinate ranges.

    Supports visualization of kernel functions with different dimensionalities:
    - 0D: Single point plot with error bars
    - 1D: Line plot with error bars
    - 2D: Heatmap visualization

    Attributes:
        plot: Read-only property returning the plot configuration
    """

    config_cls = KernelConfig

    @typechecked
    def __init__(
        self,
        escobj: "kernel_obj",
        coordinates: Optional[ArrayLike] = None,
        config: Optional[KernelConfig] = None,
        **kwargs,
    ):
        """Initialize the kernel layout.

        Args:
            escobj: The kernel object to visualize
            coordinates: Custom coordinate range for visualization.
                        Required for n-dimensional kernels.
            config: Layout configuration options
            **kwargs: Additional keyword arguments passed to parent

        Raises:
            ValueError: If number of variables is not 1 or 2
            ValueError: If coordinates array length is not a multiple of dimensions
        """

        dims = escobj.num_variables
        if dims > 2:
            raise ValueError(
                f"Unexpected number of variables, expected one of [1, 2], got {dims}"
            )

        if coordinates is None:
            if dims <= 1:  # Handle both 0D and 1D cases
                coordinates = np.linspace(-10, 10, 300, dtype=float).copy()
            else:
                x = np.linspace(-10, 10, 300, dtype=float)
                y = np.linspace(-10, 10, 300, dtype=float)
                grid = np.meshgrid(x, y)
                # flatten grid arrays
                xvf = grid[0].flatten()
                yvf = grid[1].flatten()
                # create coordinates array of type [[x0, y0], [x1, y1],...]
                coordinates = np.column_stack([xvf, yvf]).flatten()
        else:
            coordinates = np.array(coordinates, dtype=float).flatten()
            if len(coordinates) % dims != 0:
                raise ValueError(
                    "Coordinates array length must be a multiple of the number of variables"
                )

        if config is None:
            config = KernelConfig()
        self._coordinates = coordinates
        super(KernelObjectLayout, self).__init__(escobj, config=config, **kwargs)

    def prepare_data(self) -> tuple[np.ndarray, Optional[None], Optional[tuple]]:
        """Prepare kernel data for visualization.

        Returns:
            tuple containing:
                - Array of kernel values
                - None (placeholder for error values)
                - Optional extent tuple for 2D plots
        """
        escobj = self._escobj
        num_variables: int = escobj.num_variables
        xarr: np.ndarray = self._coordinates
        yarr = np.zeros(shape=(xarr.size // num_variables,))
        escobj(xarr, yarr)

        if num_variables <= 1:
            return xarr, yarr, None
        elif num_variables == 2:
            x0arr = xarr[0::num_variables]
            x1arr = xarr[1::num_variables]

            # Check if data is on a regular grid by comparing unique coordinates
            x0_unique = unique(x0arr, tol=1e-8)
            x1_unique = unique(x1arr, tol=1e-8)

            # Count points that should be in a regular grid
            expected_points = len(x0_unique) * len(x1_unique)
            actual_points = len(yarr)

            x0_min, x0_max = np.min(x0arr), np.max(x0arr)
            x1_min, x1_max = np.min(x1arr), np.max(x1arr)

            if expected_points != actual_points:
                # Data is irregular - need interpolation
                from scipy.interpolate import griddata

                # Estimate grid size based on point density
                grid_size = int(np.sqrt(len(x0arr)))
                x0_grid = np.linspace(x0_min, x0_max, grid_size)
                x1_grid = np.linspace(x1_min, x1_max, grid_size)
                X0, X1 = np.meshgrid(x0_grid, x1_grid)

                # Interpolate data onto regular grid
                points = np.column_stack((x0arr, x1arr))
                yarr_interp = griddata(points, yarr, (X0, X1), method="cubic")

                # Update arrays with interpolated values
                yarr = yarr_interp
                extent = (x0_min, x0_max, x1_min, x1_max)

            else:
                # Data is already on regular grid - reshape directly
                rows = len(x1_unique)
                cols = len(x0_unique)

                yarr = yarr.reshape((rows, cols))

                extent = (x0_min, x0_max, x1_min, x1_max)

            return yarr, None, extent
