from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional, Union
from typing_extensions import Self

import numpy as np
from escape.utils.widgets.base import BaseLayoutWithPlot, BaseStackLayout, LayoutConfig
from escape.utils.widgets.plot import (
    MapPlotConfig,
    ModelMapWidget,
    ModelPlot2dWidget,
    Plot2dConfig,
)
from typeguard import typechecked

if TYPE_CHECKING:
    from escape.core.model import model_obj, modelstack_obj


@dataclass(frozen=True)
class ModelConfig(LayoutConfig, Plot2dConfig, MapPlotConfig):
    """Configuration for model object layout."""


class ModelObjectLayout(BaseLayoutWithPlot):
    """Layout for displaying model objects with data, simulation and residuals.

    Supports visualization of model data with different dimensionalities:
    - 0D/1D: Line plots with error bars, simulation curves and residuals
    - 2D: Heatmap visualizations with simulation and residual maps

    Attributes:
        plot: Read-only property returning the plot configuration
    """

    config_cls = ModelConfig
    plot2d_cls = ModelPlot2dWidget
    map_cls = ModelMapWidget

    @typechecked
    def __init__(
        self, escobj: "model_obj", config: Optional[ModelConfig] = None, **kwargs
    ):
        """Initialize the model layout.

        Args:
            escobj: The model object to visualize
            config: Layout configuration options
            **kwargs: Additional keyword arguments passed to parent

        Raises:
            ValueError: If number of variables is not 0, 1 or 2
        """
        if config is None:
            config = ModelConfig()
        dims = escobj.num_variables
        if dims > 2:
            raise ValueError(
                f"Unexpected number of variables, expected one of [0, 1, 2], got {dims}"
            )
        super(ModelObjectLayout, self).__init__(escobj, config=config, **kwargs)

    def prepare_data(self):
        """Prepare model data for visualization.

        Processes model data, simulation and residuals into appropriate format
        based on dimensionality.

        Returns:
            For 0D/1D:
                tuple containing:
                - Masked array of x values
                - Masked array of y values
                - Masked array of error values
                - Masked array of simulation values
                - Masked array of residual values
                - Model name
                - Data name

            For 2D:
                tuple containing:
                - 2D masked array of y values
                - 2D masked array of simulation values
                - 2D masked array of residual values
                - Extent tuple (x0_min, x0_max, x1_min, x1_max)

        Raises:
            ValueError: If number of variables is not 0, 1 or 2
        """
        escobj = self._escobj
        dims = escobj.num_variables

        dobj = escobj.data
        xarr = np.array(dobj.ini_coordinates)
        yarr = np.array(dobj.ini_intensities)
        earr = np.array(dobj.ini_errors) if len(dobj.ini_errors) else np.sqrt(yarr)
        marr = np.array(dobj.mask)

        escobj()
        sarr = np.array(escobj.simulation)
        rarr = np.array(escobj.residuals)

        # Create masked arrays
        masked_yarr = np.ma.masked_array(yarr, ~marr)
        masked_sarr = np.ma.masked_array(sarr, ~marr)
        masked_rarr = np.ma.masked_array(rarr, ~marr)
        masked_earr = np.ma.masked_array(earr, ~marr)

        if dims <= 1:
            # For 0D, create a single point x-coordinate
            if dims == 0:
                xarr = np.array([0.0])
            masked_xarr = np.ma.masked_array(xarr, ~marr)

            return (
                masked_xarr,
                masked_yarr,
                masked_earr,
                masked_sarr,
                masked_rarr,
                escobj.name,
                dobj.name,
            )

        elif dims == 2:
            # Handle 2D data
            x0arr = xarr[0::dims]
            x1arr = xarr[1::dims]

            # Get unique coordinates to determine grid shape
            x0_unique = np.unique(x0arr)
            x1_unique = np.unique(x1arr)

            rows = len(x1_unique)
            cols = len(x0_unique)

            # Reshape arrays to 2D
            yarr_2d = masked_yarr.reshape((rows, cols))
            sarr_2d = masked_sarr.reshape((rows, cols))
            rarr_2d = masked_rarr.reshape((rows, cols))

            # Calculate extent
            extent = (np.min(x0arr), np.max(x0arr), np.min(x1arr), np.max(x1arr))

            return (yarr_2d, sarr_2d, rarr_2d, extent)

        else:
            raise ValueError(f"Unexpected number of variables: {dims}")


class ModelStackLayout(BaseStackLayout):
    """Layout for displaying a stack of model objects.

    Supports visualization of multiple model objects with different dimensionalities:
    - 0D/1D: Line plots with error bars, simulation curves and residuals
    - 2D: Heatmap visualizations with simulation and residual maps
    """

    item_config_cls = ModelConfig

    @typechecked
    def __init__(
        self,
        model_stack: "modelstack_obj",
        model_configs: Optional[Union[ModelConfig, list[ModelConfig]]] = None,
        config: Optional[LayoutConfig] = None,
        **kwargs,
    ) -> None:
        """Initialize the model stack layout.

        Args:
            model_stack: Model stack object containing multiple models to visualize
            model_configs: Configuration options for each model object
            config: Layout configuration options
            **kwargs: Additional keyword arguments passed to parent
        """
        escobjs = [model_stack.model(i) for i in range(len(model_stack))]
        super().__init__(escobjs, object_configs=model_configs, config=config, **kwargs)

    def config_model(
        self,
        index: Optional[int] = None,
        title: Union[str, list[str], tuple[str, ...]] = "",
        xlabel: Union[str, list[str], tuple[str, ...]] = "",
        ylabel: Union[str, list[str], tuple[str, ...]] = "",
        xlog: Union[bool, list[bool], tuple[bool, ...]] = False,
        ylog: Union[bool, list[bool], tuple[bool, ...]] = False,
        grid: Union[bool, list[bool], tuple[bool, ...]] = False,
        cbmin: Union[
            Optional[float], list[Optional[float]], tuple[Optional[float], ...]
        ] = None,
        cbmax: Union[
            Optional[float], list[Optional[float]], tuple[Optional[float], ...]
        ] = None,
        cblog: Union[bool, list[bool], tuple[bool, ...]] = False,
        colorscale: Union[str, list[str], tuple[str, ...]] = "Viridis",
        fig_title: Union[str, list[str], tuple[str, ...]] = "",
        data_title: Union[str, list[str], tuple[str, ...]] = "",
        error_title: Union[str, list[str], tuple[str, ...]] = "",
        show_errors: Union[bool, list[bool], tuple[bool, ...]] = True,
    ) -> Self:
        """Configure the layout with a title and maximum number of columns.

        Args:
            index: Index of the model object to configure. If None, all models will be configured.
            title: Title to display in header. Can be str or list/tuple[str].
            xlabel: Label for x-axis. Can be str or list/tuple[str].
            ylabel: Label for y-axis. Can be str or list/tuple[str].
            xlog: Scale for x-axis. Can be bool or list/tuple[bool].
            ylog: Scale for y-axis. Can be bool or list/tuple[bool].
            grid: Whether to show grid. Can be bool or list/tuple[bool].
            cbmin: Min value for colorbar. Can be float or list/tuple[float].
            cbmax: Max value for colorbar. Can be float or list/tuple[float].
            cblog: Log scale for colorbar. Can be bool or list/tuple[bool].
            colorscale: Colormap name. Can be str or list/tuple[str].
            fig_title: Title for figure. Can be str or list/tuple[str].
            data_title: Title for data. Can be str or list/tuple[str].
            error_title: Title for error. Can be str or list/tuple[str].
            show_errors: Whether to show errors. Can be bool or list/tuple[bool].
        """

        def _setup_config(model_widget: ModelObjectLayout, idx: int):
            config = self.item_config_cls(
                title=title[idx] if isinstance(title, (list, tuple)) else title,
                xlabel=xlabel[idx] if isinstance(xlabel, (list, tuple)) else xlabel,
                ylabel=ylabel[idx] if isinstance(ylabel, (list, tuple)) else ylabel,
                xlog=xlog[idx] if isinstance(xlog, (list, tuple)) else xlog,
                ylog=ylog[idx] if isinstance(ylog, (list, tuple)) else ylog,
                grid=grid[idx] if isinstance(grid, (list, tuple)) else grid,
                cbmin=cbmin[idx] if isinstance(cbmin, (list, tuple)) else cbmin,
                cbmax=cbmax[idx] if isinstance(cbmax, (list, tuple)) else cbmax,
                cblog=cblog[idx] if isinstance(cblog, (list, tuple)) else cblog,
                colorscale=(
                    colorscale[idx]
                    if isinstance(colorscale, (list, tuple))
                    else colorscale
                ),
                fig_title=(
                    fig_title[idx]
                    if isinstance(fig_title, (list, tuple))
                    else fig_title
                ),
                data_title=(
                    data_title[idx]
                    if isinstance(data_title, (list, tuple))
                    else data_title
                ),
                error_title=(
                    error_title[idx]
                    if isinstance(error_title, (list, tuple))
                    else error_title
                ),
                show_errors=(
                    show_errors[idx]
                    if isinstance(show_errors, (list, tuple))
                    else show_errors
                ),
            )
            model_widget.cleanup()
            model_widget._config = config
            model_widget.init_layout()

        if index is None:
            # Validate list lengths match number of models if lists provided
            num_models = len(self._widgets)
            for param_name, param_val in locals().items():
                if isinstance(param_val, (list, tuple)) and param_name != "self":
                    if len(param_val) != num_models:
                        raise ValueError(
                            f"{param_name} list length {len(param_val)} does not match number of models {num_models}"
                        )

            for i, model_widget in enumerate(self._widgets):
                _setup_config(model_widget, i)
        else:
            model_widget = self._widgets[index]
            _setup_config(model_widget, 0)
        return self
