from dataclasses import asdict, dataclass
from typing import TYPE_CHECKING, Optional

import numpy as np
from typeguard import typechecked

if TYPE_CHECKING:
    from escape.core.multilayer import multilayer_obj
    from escape.core.source import source_obj

from escape.utils.widgets.kernel import KernelConfig, KernelObjectLayout
from escape.utils.widgets.plot import Plot2dConfig

SLD0RE_TYPE = "sld0re"
SLD0IM_TYPE = "sld0im"
SLDM_TYPE = "sldm"


@dataclass(frozen=True)
class ProfileConfig(KernelConfig):
    """Configuration for profile object layout."""


class ProfileLayout(KernelObjectLayout):
    """Layout for displaying SLD profiles of multilayer objects.

    Provides visualization of scattering length density (SLD) profiles for multilayer
    structures, supporting real and imaginary components.

    Attributes:
        plot: Read-only property returning the plot configuration
    """

    config_cls = ProfileConfig

    @typechecked
    def __init__(
        self,
        escobj: "multilayer_obj",
        profile: str = SLD0RE_TYPE,
        source: Optional["source_obj"] = None,
        config: Optional[ProfileConfig] = None,
        **kwargs,
    ):
        """Initialize profile visualization.

        Args:
            escobj: Multilayer object to visualize
            title: Title to display in header
            profile: Type of profile to display (sld0re, sld0im, or sldm)
            source: Optional source object for units
            layout: Layout configuration options
            plot: Plot configuration options
            **kwargs: Additional keyword arguments passed to parent

        Raises:
            TypeError: If escobj is not a multilayer object
            ValueError: If profile type is invalid or minimum thickness is 0
        """
        from escape.core.kernel import kernel

        # Create profile function
        func = self._create_profile_function(profile, escobj, source)

        # Set up units
        units = self._get_units(source)

        # Create kernel
        k = kernel(escobj.name, func, multithreaded=True)

        # Handle coordinates
        coordinates = self._setup_coordinates(escobj)

        # Configure plot
        if config is None:
            config = ProfileConfig()

        config_dict = asdict(config)
        config_dict["xlabel"] = f"Z[{units}]"
        config_dict["ylabel"] = f"{profile.upper()}[1/{units}²]"
        if not config_dict["title"] is None:
            config_dict["title"] = f"{escobj.name} [{profile.upper()}]"

        _config = KernelConfig(**config_dict)
        super(ProfileLayout, self).__init__(
            k, coordinates=coordinates, config=_config, **kwargs
        )

    def _create_profile_function(
        self, profile: str, escobj: "multilayer_obj", src: Optional["source_obj"]
    ):
        """Create the profile function based on type.

        Args:
            profile: Type of profile to create
            escobj: Multilayer object
            src: Source object for units

        Returns:
            Profile function

        Raises:
            ValueError: If profile type is invalid
            NotImplementedError: If SLDM profile is requested
        """
        from escape import imag, profile_sld0, profile_sldm, real, var

        if profile == SLD0RE_TYPE:
            prof = profile_sld0("", var("z"), escobj, src)
            return real(prof)
        elif profile == SLD0IM_TYPE:
            prof = profile_sld0("", var("z"), escobj, src)
            return imag(prof)
        elif profile == SLDM_TYPE:
            prof = profile_sldm("", var("z"), escobj, src)
            return prof
        else:
            raise ValueError(f"Invalid profile type: {profile}")

    def _get_units(self, source: Optional["source_obj"]) -> str:
        """Determine units based on source.

        Args:
            source: Source object containing units information

        Returns:
            Unit string (angstr or nm)
        """
        from escape import angstr

        if source is None:
            return ""
        return angstr if source.units.lower() == "angstrom" else "nm"

    def _setup_coordinates(self, escobj: "multilayer_obj") -> np.ndarray:
        """Set up coordinate range for profile.

        Args:
            escobj: Multilayer object

        Returns:
            Array of z coordinates

        Raises:
            ValueError: If minimum thickness is 0
        """
        zmax = escobj.total_thickness * 1.5
        if zmax == 0:  # substrate only
            zmax = 200
            numpoints = 100
        else:
            minthkn = escobj.min_thickness
            if minthkn == 0:
                raise ValueError("Minimum thickness cannot be 0")
            numpoints = max(int(zmax / minthkn) * 5, 50)

        return np.linspace(-min(100, zmax), zmax, numpoints)
