import io
import threading
import time
import traceback
from typing import TYPE_CHECKING, Optional, Union

import ipywidgets.widgets as widgets
import numpy as np
from escape.utils.widgets.base import LayoutConfig
from escape.utils.widgets.model import ModelConfig, ModelStackLayout
from escape.utils.widgets.plot import Plot2dConfig, Plot2dWidget
from typeguard import typechecked

if TYPE_CHECKING:
    from escape.utils.regressor.base import BaseRegressor


class RegressorObjectLayout(ModelStackLayout):
    """Layout for regressor visualization and control.

    Provides interactive controls for training/testing the regressor,
    parameter manipulation, and visualization of training progress.

    Attributes:
        plot: Read-only property returning the plot configuration
    """

    center_height: int = 600
    header_height: int = 50
    footer_height: int = 180


    @typechecked
    def __init__(
        self,
        regressor: "BaseRegressor",
        model_configs: Optional[
            Union[ModelConfig, list[ModelConfig], tuple[ModelConfig, ...]]
        ] = None,
        config: Optional[LayoutConfig] = None,
        **kwargs,
    ):
        """Initialize the regressor object layout.

        Args:
            regressor: The regressor object to visualize
            model_configs: Configuration options for model visualization
            config: Layout configuration options
            **kwargs: Additional keyword arguments passed to parent
        """
        if config is None:
            config = LayoutConfig()
        self._regressor = regressor
        modelstack = regressor.modelstack
        super().__init__(
            modelstack, model_configs=model_configs, config=config, **kwargs
        )

    def _update_train_button(self, is_training: bool):
        """Update the train button appearance based on training state."""
        if is_training:
            self._trainbtn.description = "⏳"
            self._trainbtn.tooltip = "Training in progress..."
        else:
            self._trainbtn.description = "Train"
            self._trainbtn.tooltip = "Start training"

    def on_train_begin(self, net, X=None, y=None, **kwargs):
        """Handle start of training."""
        self._status_label.value = "<b>Training...</b>"
        self._update_train_button(True)

    def on_train_end(self, net, X=None, y=None, **kwargs):
        """Handle end of training."""
        self._status_label.value = "<b>Done</b>"
        self._update_train_button(False)
        escobj = self._regressor
        escobj.on_train_begin = None
        escobj.on_train_end = None
        escobj.on_epoch_begin = None
        escobj.on_epoch_end = None

    def on_epoch_begin(self, net, dataset_train=None, dataset_valid=None, **kwargs):
        """Handle start of training epoch."""
        pass

    def on_epoch_end(self, net, dataset_train=None, dataset_valid=None, **kwargs):
        """Handle end of training epoch and update progress plot."""
        try:
            self.update_loss_plot()
        except Exception as err:
            self._status_label.value = f"<b>Error:</b> {err}"

    def update_status(self):
        """Update training status display."""
        try:
            train_score = self._regressor.score(
                self._regressor.X_train, self._regressor.y_train
            )
            test_score = self._regressor.score(
                self._regressor.X_test, self._regressor.y_test
            )
            self._status_label.value = (
                f"Training Score: {train_score:.4g} " f"Test Score: {test_score:.4g} "
            )
        except Exception as e:
            print(f"Error updating status: {e}")

    def init_header(self, config: Optional[LayoutConfig] = None):
        """Initialize header with training control buttons."""

        def on_train_click(btn):
            self._disable_controls(True)
            try:
                escobj = self._regressor
                #escobj.on_train_begin = self.on_train_begin
                #escobj.on_train_end = self.on_train_end
                #escobj.on_epoch_begin = self.on_epoch_begin
                escobj.on_epoch_end = self.on_epoch_end
                self._update_train_button(True)
                def train_thread():
                    try:
                        self._status_label.value = f"<b>Training...</b>"
                        score = escobj.train(True)
                        self._status_label.value = f"<b>Training complete. Score:</b> {score:.4g}"
                        self.update_parameter_sliders(silent=True)
                        self.update()
                        self._disable_controls(False)
                        self._update_train_button(False)
                    except Exception as err:
                        if self._regressor.is_stopped:
                            self._status_label.value = f"<b>Training stopped</b>"
                        else:
                            self._status_label.value = f"<b>Error:</b> {err}"
                        self._update_train_button(False)
                        self._disable_controls(False)
                threading.Thread(target=train_thread).start()
                
            except Exception as err:
                self._status_label.value = f"<b>Error:</b> {err}"
                self._update_train_button(False)
                self._disable_controls(False)

        def on_stop_click(btn):
            # Implement early stopping mechanism
            threading.Thread(target=self._regressor.stop).start()

        def on_reset_click(btn):
            # Reset model parameters
            try:
                self._regressor.reset()
                self.update_parameter_sliders(silent=True)
                self.update()
                self.update_status()
            except Exception as err:
                self.status_label.value = f"<b>Error:</b> {err}"
                traceback.print_exc()

        def on_test_click(btn):
            try:
                data = self._regressor.test()
                self._regressor(data)
                self.update_parameter_sliders(silent=True)
                self.update()
            except Exception as err:
                self._status_label.value = f"<b>Error:</b> {err}"
                traceback.print_exc()

        btnlay = widgets.Layout(width="60px")

        self._trainbtn = widgets.Button(
            disabled=False, tooltip="Start training", description="Train", layout=btnlay
        )
        self._trainbtn.on_click(on_train_click)

        self._stopbtn = widgets.Button(
            disabled=True, tooltip="Stop training", description="Stop", layout=btnlay
        )
        self._stopbtn.on_click(on_stop_click)

        self._testbtn = widgets.Button(
            disabled=False, tooltip="Test model", description="Test", layout=btnlay
        )
        self._testbtn.on_click(on_test_click)

        self._resetbtn = widgets.Button(
            disabled=False,
            tooltip="Reset model parameters",
            description="Reset",
            layout=btnlay,
        )
        self._resetbtn.on_click(on_reset_click)

        toolbar = widgets.HBox(
            children=[self._trainbtn, self._stopbtn, self._testbtn, self._resetbtn]
        )
        return toolbar

    def init_footer(self, config: Optional[LayoutConfig] = None):
        """Initialize footer with loss plot and status label."""
        plot = Plot2dConfig(xlabel="Epochs", ylabel="Loss")
        self._loss_plot = Plot2dWidget(
            self, config=plot, height=int(self.pane_heights[2][:-2])-50#50 is for status_label
        )
        self._status_label = widgets.HTML("Status: ")
        return widgets.VBox([self._status_label, self._loss_plot])

    def update_loss_plot(self):
        """Update the loss plot."""
        if (
            hasattr(self._regressor, "history")
            and self._regressor.history is not None
        ):
            try:
                losses = self._regressor.history[:, "train_loss"]
                epochs = np.arange(len(losses))
                self._loss_plot.update(epochs, losses)
            except KeyError as err: #train_loss is not available
                pass

    def _disable_controls(self, disabled: bool):
        """Enable/disable all control buttons.

        Args:
            disabled: Whether to disable the controls
        """
        self._trainbtn.disabled = disabled
        self._resetbtn.disabled = disabled
        self._testbtn.disabled = disabled
        self._stopbtn.disabled = not disabled
        self.disable_parameters(disabled)
